#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <math.h>
#include <string.h>
#include "op.h"
#include "type.h"
#include "cmd.h"

gstate_t cur_gstate = (gstate_t) { {1, 0, 0, 0, 1, 0, 0, 0, 1}, NULL, NULL, 0, 0, 0, 1 };

double devxscale = 2.0/842.0;
double devxshift = -1.0;
double devyscale = 2.0/842.0;
double devyshift = -1.0;
double minx=0, miny=0, maxx=0, maxy=0;

int beziersegs = 8;

static void pcoord(double pt[2])
{
	static int first = 1;
	double *m = cur_gstate.CTM;
	double ux = m[0] * pt[0] + m[3] * pt[1] + m[6];
	double uy = m[1] * pt[0] + m[4] * pt[1] + m[7];
	if(first) {
		minx = ux; miny = uy;
		maxx = ux; maxy = uy;
		first =0;
	} else {
		if(ux < minx) minx = ux;
		if(ux > maxx) maxx = ux;
		if(uy < miny) miny = uy;
		if(uy > maxy) maxy = uy;
	}

	printf(" %g", (ux + devxshift) * devxscale - 1);
	printf(" %g", (uy + devyshift) * devyscale - 1);
}

static int lastr=0, lastg=0, lastb=0;
static int lastwidth=1;

static void scolor()
{
	if(cur_gstate.cr != lastr || cur_gstate.cg != lastg || cur_gstate.cb != lastb) {
		lastr = cur_gstate.cr;
		lastg = cur_gstate.cg;
		lastb = cur_gstate.cb;
		printf("color #%02x%02x%02x\n", lastr, lastg, lastb);
	}

	int width = 1;
	width = (int) (cur_gstate.thick * cur_gstate.CTM[0] + 0.5);
	if(width < 1) width = 1;

	if(width != lastwidth) {
		printf("thick %d\n", width);
		lastwidth = width;
	}
}

static inline obj_t * pop()
{
	return arr_pop(opstack);
}

static inline int popint()
{
	obj_t *o = pop();
	switch(o->type) {
	case T_INT:
		return o->i;
	case T_REAL:
		return (int) o->r;
	default:
		fprintf(stderr, "top of stack not integer\n");
		abort();
	}
}

static const char * popstring()
{
	obj_t *o = pop();
	switch(o->type) {
	case T_NAME:
		return o->s;
	case T_STRING:
		return o->s;
	default:
		fprintf(stderr, "top of stack not string\n");
		abort();
	}
}

static inline int popbool()
{
	obj_t *o = pop();
	if(o->type != T_BOOL) {
		fprintf(stderr, "top of stack not bool\n");
		abort();
	}
	return o->i;
}

static inline double popreal()
{
	return getreal(pop());
}

static inline void push(obj_t *o)
{
	arr_add(opstack, o);
}

void Oprint()
{
	obj_t *o = pop();

	switch(o->type) {
	case T_NAME:
		printf("/%s", o->s);
		break;
	case T_STRING:
		printf("(%s)", o->s);
		break;
	case T_INT:
		printf("%d", o->i);
		break;
	case T_REAL:
		printf("%g", o->r);
		break;
	case T_ARRAY:
		{
			int i;
			printf("%s", o->exec ? "{" : "[");
			for(i=0; i<o->arr.size; i++) {
				printf(" ");
				arr_add(opstack, o->arr.v[i]);
				Oprint();
			}
			printf("%s", o->exec ? " }" : " ]");
		}
		break;
	case T_DICT:
		{
			printf("<<");
			ptrhash_entry_t *e;
			for(e = ptrhash_first(o->dict); e; e = ptrhash_next(o->dict, e)) {
				printf(" ");
				arr_add(opstack, e->key);
				Oprint();
				printf(" ");
				arr_add(opstack, e->value);
				Oprint();
			}

			printf(">>");
		}
		break;
	case T_MARK:
		printf("mark");
		break;
	case T_NULL:
		printf("null");
		break;
	case T_BOOL:
		printf("%s", o->i ? "true" : "false");
		break;
	case T_OP:
		printf("OP");
		break;
	case T_GSTATE:
		printf("GSTATE");
		break;
	default:
		printf("?");
		break;
	}
}

static void Ostack()
{
	printf("\n------ opstack %d els\n", opstack->arr.size);
	int i;
	for(i=0; i<opstack->arr.size; i++) {
		printf("%2d :: ", i);
		push(opstack->arr.v[i]);
		Oprint();
		printf("\n");
	}
	printf("---------\n");
}

static void Odef()
{
	obj_t *v = pop();
	obj_t *k = pop();
	dict_set(arr_peek(dictstack), k, v);
}

static void Opop()
{
	pop();
}

static void Oexch()
{
	obj_t *a = pop();
	obj_t *b = pop();
	push(a);
	push(b);
}

static void Odup()
{
	obj_t *a = pop();
	push(a);
	push(a);
}

static void Ocopy()
{
	int n = popint();
	assert(n >= 0);
	int o = opstack->arr.size - n;
	int i;
	for(i=0; i<n; i++) {
		push(arr_idx(opstack, o+i));
	}
}

static void Oindex()
{
	int n = popint();
	push(arr_idx(opstack, opstack->arr.size-1 - n));
}

static void Oclear()
{
	while(opstack->arr.size > 0) {
		pop();
	}
}

static void Oroll()
{
	int n = popint();
	int size = popint();
	assert(size <= opstack->arr.size);
	assert(size >= 0);

	if(!n) return;
	if(!size) return;

	n = 0 - (n % size);
	if(n < 0) n = size + n;

	obj_t **temp = malloc(size * sizeof *temp);
	if(!temp) abort();

	int bottom = opstack->arr.size - size;
	memcpy(temp, opstack->arr.v + bottom, size * sizeof *temp);

	int i;
	for(i=0; i<size; i++) {
		opstack->arr.v[bottom] = temp[n];
		bottom ++;
		n ++;
		if(n == size) n = 0;
	}

	free(temp);
}

static void Omark()
{
	push(mkmark());
}

static void Ocleartomark()
{
	while(pop()->type != T_MARK);
}

static void Ocounttomark()
{
	int i;
	for(i=0; arr_idx(opstack, opstack->arr.size-1 - i)->type != T_MARK; i++);
	push(mkint(i));
}

#define NUMMONOP(name, intop, realop) \
static void name () \
{ \
	obj_t *o = pop(); \
	switch(o->type) { \
	case T_INT: \
		{ \
			int i = o->i; \
			intop; \
			push(mkint(i)); \
		} \
		break; \
	case T_REAL: \
		{ \
			double r = o->r; \
			realop; \
			push(mkreal(r)); \
		} \
		break; \
	default: \
		fprintf(stderr, "invalid type for monop\n"); \
		abort(); \
	} \
}

NUMMONOP(Oabs, i=abs(i), r=fabs(r));
NUMMONOP(Oneg, i=-i, r=-r);

#define NUMBINOP(name, intop, realop) \
static void name () \
{ \
	obj_t *ob = pop(); obj_t *oa = pop(); \
	if(oa->type == T_INT && ob->type == T_INT) { \
		int a = oa->i, b = ob->i, i; \
		intop; \
		push(mkint(i)); \
	} else { \
		double a = getreal(oa), b = getreal(ob), r; \
		realop; \
		push(mkreal(r)); \
	} \
}

NUMBINOP(Oadd, i=a+b, r=a+b);
NUMBINOP(Omul, i=a*b, r=a*b);
NUMBINOP(Osub, i=a-b, r=a-b);

static void Omod()
{
	int i2 = popint();
	int i1 = popint();
	push(mkint(i1 % i2));
}

static void Odiv()
{
	double r2 = popreal();
	double r1 = popreal();
	push(mkreal(r1 / r2));
}

static void Owhere()
{
	obj_t *k = pop();
	int i;
	for(i = dictstack->arr.size - 1; i >= 0; i --) {
		obj_t *d = dictstack->arr.v[i];
		obj_t *v = dict_lookup(d, k);
		if(v) {
			push(d);
			push(mkbool(1));
			return;
		}
	}

	push(mkbool(0));
}

static void Oif()
{
	obj_t *p = pop();
	int b = popbool();
	if(b) execobj(p);
}

static void Oifelse()
{
	obj_t *p2 = pop();
	obj_t *p1 = pop();
	int b = popbool();
	obj_t *e = b ? p1 : p2;
	execobj(e);
}

#define CMPOP(name, ltres, eqres, gtres) \
static void name () \
{ \
	obj_t *o2 = pop(); obj_t *o1 = pop(); \
	objcmp_t c = obj_cmp(o1, o2); \
	switch(c) { \
	case C_LT: push(mkbool(ltres)); break; \
	case C_EQ: push(mkbool(eqres)); break; \
	case C_GT: push(mkbool(gtres)); break; \
	case C_INVALID: abort(); \
	} \
}

CMPOP(Ole, 1, 1, 0);
CMPOP(Olt, 1, 0, 0);
CMPOP(Oge, 0, 1, 1);
CMPOP(Ogt, 0, 0, 1);

static void Oeq()
{
	obj_t *b = pop();
	obj_t *a = pop();
	if((a->type == T_INT && b->type == T_REAL) || (a->type == T_REAL && b->type == T_INT)) {
		push(mkbool(getreal(a) == getreal(b)));
		return;
	}
	push(mkbool(obj_hash_eq(a, b)));
}

static void One()
{
	obj_t *b = pop();
	obj_t *a = pop();
	if((a->type == T_INT && b->type == T_REAL) || (a->type == T_REAL && b->type == T_INT)) {
		push(mkbool(getreal(a) != getreal(b)));
		return;
	}
	push(mkbool(!obj_hash_eq(a, b)));
}

#define NOTIMPL(name, args) \
static void O ## name () \
{ \
	int i; \
	for(i=0; i<args; i++) pop(); \
	fprintf(stderr, "Warning: %s not implemented\n", #name); \
}

NOTIMPL(findfont, 1);
NOTIMPL(scalefont, 1);
NOTIMPL(setfont, 0);
NOTIMPL(bind, 0);
NOTIMPL(setlinecap, 1);
NOTIMPL(setlinejoin, 1);
NOTIMPL(setdash, 2);
NOTIMPL(setmiterlimit, 1);
NOTIMPL(rectclip, 4);
NOTIMPL(setcolorspace, 1);
NOTIMPL(clip, 0);

static void Obegin()
{
	obj_t *d = pop();
	assert(d->type == T_DICT);
	arr_add(dictstack, d);
}

static void Oend()
{
	arr_pop(dictstack);
}

static void Oload()
{
	obj_t *k = pop();
	int i;
	for(i = dictstack->arr.size - 1; i >= 0; i --) {
		obj_t *d = dictstack->arr.v[i];
		obj_t *v = dict_lookup(d, k);
		if(v) {
			push(v);
			return;
		}
	}
	fprintf(stderr, "load didn't find key\n");
	abort();
}

static void Ogsave()
{
	arr_add(gstatestack, mkgstate(&cur_gstate));
}

static void Ogrestore()
{
	obj_t *g = arr_pop(gstatestack);
	assert(g->type == T_GSTATE);
	memcpy(&cur_gstate, g->gs, sizeof(gstate_t));
}

static inline void color_limit()
{
#define LIMITBYTE(c) if(c < 0) c = 0; else if(c > 255) c = 255;
	LIMITBYTE(cur_gstate.cr);
	LIMITBYTE(cur_gstate.cg);
	LIMITBYTE(cur_gstate.cb);
}

static void Osetgray()
{
	double g = popreal();
	int l = (int) (g * 255.0);
	cur_gstate.cr = l;
	cur_gstate.cg = l;
	cur_gstate.cb = l;
	color_limit();
}

static void Osetrgbcolor()
{
	double b = popreal();
	double g = popreal();
	double r = popreal();
	cur_gstate.cr = (int) (r * 255.0);
	cur_gstate.cg = (int) (g * 255.0);
	cur_gstate.cb = (int) (b * 255.0);
	color_limit();
}

static void Omoveto()
{
	double y = popreal();
	double x = popreal();
	path_t *p = cur_gstate.path;

	if(p->wasmoveto) {
		assert(!p->closed);
		spath_t *l = path_last(p);
		assert(l);
		double *p = spath_pt(l, l->cnt-1);
		p[0] = x;
		p[1] = y;
	} else {
		path_newsub(p);
		p->wasmoveto = 1;
		p->closed = 0;
		spath_t *l = path_last(p);
		spath_add(l, x, y);
	}
}

static void Olineto()
{
	double y = popreal();
	double x = popreal();

	path_t *p = cur_gstate.path;
	p->wasmoveto = 0;
	spath_t *l = path_last(p);

	if(p->closed) {
		assert(l->cnt > 0);
		double *op = spath_last(l);
		path_newsub(p);
		l = path_last(p);
		spath_add(l, op[0], op[1]);
	}

	assert(l);
	assert(l->cnt > 0);
	spath_add(l, x, y);
}

static void Orlineto()
{
	double dy = popreal();
	double dx = popreal();

	path_t *p = cur_gstate.path;
	p->wasmoveto = 0;
	spath_t *l = path_last(p);

	if(p->closed) {
		assert(l->cnt > 0);
		double *op = spath_last(l);
		path_newsub(p);
		l = path_last(p);
		spath_add(l, op[0], op[1]);
	}

	assert(l);
	assert(l->cnt > 0);
	double *p0 = spath_last(l);

	spath_add(l, p0[0] + dx, p0[1] + dy);
}

static void Ocurveto()
{
	double p3[2];
	double p2[2];
	double p1[2];

	p3[1] = popreal();
	p3[0] = popreal();
	p2[1] = popreal();
	p2[0] = popreal();
	p1[1] = popreal();
	p1[0] = popreal();

	path_t *p = cur_gstate.path;
	p->wasmoveto = 0;
	spath_t *l = path_last(p);
	if(p->closed) {
		assert(l->cnt > 0);
		double *op = spath_last(l);
		path_newsub(p);
		l = path_last(p);
		spath_add(l, op[0], op[1]);
	}

	assert(l);
	assert(l->cnt > 0);

	double *p0 = spath_last(l);
	double step = 1.0 / (double) beziersegs;
	double a = step;
	int i;

	double pw3(double a) {
		return a*a*a;
	}
	double pw2(double a) {
		return a*a;
	}
	double sum[2];
	void addscale(double *pt, double x) {
		sum[0] = sum[0] + pt[0] * x;
		sum[1] = sum[1] + pt[1] * x;
	}

	for(i=0; i<beziersegs; i++, a+=step) {
		sum[0] = 0; sum[1] = 0;
		addscale(p0, pw3(1-a));
		addscale(p1, 3*a*pw2(1-a));
		addscale(p2, 3*a*a*(1-a));
		addscale(p3, a*a*a);
		spath_add(l, sum[0], sum[1]);
	}
}

static void Oclosepath()
{
	path_t *p = cur_gstate.path;
	p->wasmoveto = 0;
	spath_t *l = path_last(p);
	assert(l);
	if(!l->cnt) return;
	if(p->closed) return;
	double *op = spath_pt(l, 0);
	spath_add(l, op[0], op[1]);
	p->closed = 1;
}

static void Onewpath()
{
	cur_gstate.path = path_init();
}

static void Ofill()
{
	fprintf(stderr, "Warning: non-zero rule fill\n");
	scolor();
	path_t *p = cur_gstate.path;
	int i, j;
	int pp=0;
	int havelast=0;
	double *last;

	for(i=0; i<p->cnt; i++) {
		spath_t *sp = p->sp[i];
		double *pt = sp->data;
		/*if(sp->cnt < 3) {
			fprintf(stderr, "Warning: subpath with < 3 points\n");
			continue;
		} */
		if(!pp) {
			printf("apoly");
			pp=1;
		}

		for(j=0; j<sp->cnt; j++, pt+=2) {
			pcoord(pt);
		}

		if(havelast) {
			pcoord(sp->data);
			pcoord(last);
		} else {
			last = sp->data;
			pcoord(last);
			havelast = 1;
		}
	}
	if(pp) {
		printf("\n");
	}

	cur_gstate.path = path_init();
}

static void Oeofill()
{
	scolor();
	path_t *p = cur_gstate.path;
	int i, j;
	int pp=0;
	int havelast=0;
	double *last;

	for(i=0; i<p->cnt; i++) {
		spath_t *sp = p->sp[i];
		double *pt = sp->data;
		/*if(sp->cnt < 3) {
			fprintf(stderr, "Warning: subpath with < 3 points\n");
			continue;
		} */
		if(!pp) {
			printf("apoly");
			pp=1;
		}

		for(j=0; j<sp->cnt; j++, pt+=2) {
			//if(j == sp->cnt - 1 && j > 0 && pt[0] == sp->data[0] && pt[1] == sp->data[1]) continue;
			pcoord(pt);
		}

		if(havelast) {
			pcoord(sp->data);
			pcoord(last);
		} else {
			last = sp->data;
			pcoord(last);
			havelast = 1;
		}
	}
	if(pp) {
		printf("\n");
	}

	cur_gstate.path = path_init();
}

static void Ostroke()
{
	scolor();

	path_t *p = cur_gstate.path;
	int i, j;

	for(i=0; i<p->cnt; i++) {
		spath_t *sp = p->sp[i];
		double *pt = sp->data;
		/*if(sp->cnt < 2) {
			fprintf(stderr, "Warning: subpath with <2 points\n");
			continue;
		} */

		printf("lines");
		for(j=0; j<sp->cnt; j++, pt+=2) {
			pcoord(pt);
		}
		printf("\n");
	}
}

static void Oshowpage()
{
	printf("flush\n");
}

static void Osetlinewidth()
{
	double l = popreal();
	cur_gstate.thick = l;
}

static void Oarrbegin()
{
	push(mkmark());
}

static void Oarrend()
{
	int i;
	for(i=0; arr_idx(opstack, opstack->arr.size-1 - i)->type != T_MARK; i++);
	int j;
	obj_t *a = mkarr();
	for(j=opstack->arr.size-i; j<opstack->arr.size; j++) {
		arr_add(a, opstack->arr.v[j]);
	}
	for(j=0; j<=i; j++) pop();
	push(a);
}

static void Oarray()
{
	int i = popint();
	assert(i>=0);
	obj_t *a = mkarr();
	int j;
	for(j=0; j<i; j++) arr_add(a, mknull());
	push(a);
}

static void Odict()
{
	popint();
	push(mkdict());
}

static void Oastore()
{
	obj_t *a = pop();
	assert(a->type = T_ARRAY);
	int i;
	for(i=a->arr.size-1; i>=0; i--) {
		a->arr.v[i] = pop();
	}
	push(a);
}

static void Oconcat()
{
	obj_t *a = pop();
	assert(a->type = T_ARRAY);
	assert(a->arr.size == 6);


	double bm[9] = { getreal(a->arr.v[0]), getreal(a->arr.v[1]), 0,
		getreal(a->arr.v[2]), getreal(a->arr.v[3]), 0,
		getreal(a->arr.v[4]), getreal(a->arr.v[5]), 1 };

	double *am = cur_gstate.CTM;
	double rm[9];

	rm[0] = am[0] * bm[0] + am[1] * bm[3] + am[2] * bm[6];
	rm[1] = am[0] * bm[1] + am[1] * bm[4] + am[2] * bm[7];
	rm[2] = am[0] * bm[2] + am[1] * bm[5] + am[2] * bm[8];

	rm[3] = am[3] * bm[0] + am[4] * bm[3] + am[5] * bm[6];
	rm[4] = am[3] * bm[1] + am[4] * bm[4] + am[5] * bm[7];
	rm[5] = am[3] * bm[2] + am[4] * bm[5] + am[5] * bm[8];

	rm[6] = am[6] * bm[0] + am[7] * bm[3] + am[8] * bm[6];
	rm[7] = am[6] * bm[1] + am[7] * bm[4] + am[8] * bm[7];
	rm[8] = am[6] * bm[2] + am[7] * bm[5] + am[8] * bm[8];

	memcpy(cur_gstate.CTM, rm, sizeof(rm));
}

static void Oshow()
{
	const char *s = popstring();

	path_t *p = cur_gstate.path;
	spath_t *sp = path_last(p);
	assert(sp);

	printf("text right top ");
	pcoord(spath_last(sp));
	printf(" \"");
	const char *c;
	for(c = s; *c; c++) {
		switch(*c) {
		case '\n':
			printf("\\n");
			break;
		case '\"':
			printf("\\\"");
			break;
		default:
			putchar(*c);
		}
	}
	printf("\"\n");
}

static void Oquit()
{
	exit(0);
}

void op_init(obj_t *dict)
{
	cur_gstate.path = path_init();
	cur_gstate.clippath = path_init();

	dict_set(dict, mkname("["), mkop(Oarrbegin));
	dict_set(dict, mkname("]"), mkop(Oarrend));

#define OP(k) dict_set(dict, mkname(#k), mkop(O ## k));
	OP(print);
	OP(def);
	OP(pop);
	OP(exch);
	OP(dup);
	OP(copy);
	OP(stack);
	OP(index);
	OP(clear);
	OP(roll);

	OP(mark);
	OP(cleartomark);
	OP(counttomark);

	OP(array);
	OP(astore);
	OP(dict);

	OP(concat);

	OP(abs);
	OP(neg);
	OP(add);
	OP(sub);
	OP(mul);
	OP(div);
	OP(mod);

	OP(le);
	OP(lt);
	OP(ge);
	OP(gt);
	OP(eq);
	OP(ne);

	OP(where);
	OP(ifelse);
	OP(if);

	OP(bind);
	OP(begin);
	OP(end);
	OP(load);

	OP(findfont);
	OP(scalefont);
	OP(setfont);

	OP(setgray);
	OP(setrgbcolor);
	OP(setlinewidth);
	OP(setlinecap);
	OP(setlinejoin);
	OP(setdash);
	OP(setmiterlimit);

	OP(gsave);
	OP(grestore);

	OP(moveto);
	OP(lineto);
	OP(rlineto);
	OP(curveto);
	OP(closepath);
	OP(newpath);

	OP(fill);
	OP(eofill);
	OP(stroke);
	OP(showpage);
	OP(show);
	OP(quit);

	OP(rectclip);
	OP(setcolorspace);
	OP(clip);
}

