#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "type.h"
#include "gc.h"

obj_t *mkobj(type_t type)
{
	obj_t *v = GC_malloc(sizeof *v);
	if(!v) abort();
	v->type = type;
	return v;
}

obj_t *mkint(int v)
{
	obj_t *o = mkobj(T_INT);
	o->i = v;
	return o;
}

obj_t *mkreal(double r)
{
	obj_t *o = mkobj(T_REAL);
	o->r = r;
	return o;
}

obj_t *mkbool(int v)
{
	obj_t *o = mkobj(T_BOOL);
	o->i = v;
	return o;
}

obj_t *mkarr()
{
	return mkobj(T_ARRAY);
}

obj_t *mknull()
{
	return mkobj(T_NULL);
}

obj_t *mkmark()
{
	return mkobj(T_MARK);
}

obj_t *mkdict()
{
	obj_t *o = mkobj(T_DICT);
	o->dict = ptrhash_alloc((ptrhash_hash_t (*) (ptrhash_key_t)) obj_hash, (int (*) (ptrhash_key_t, ptrhash_key_t)) obj_hash_eq);
	return o;
}

obj_t *mkstring(const char *s)
{
	obj_t *o = mkobj(T_STRING);
	o->s = GC_malloc_atomic(strlen(s) + 1);
	strcpy(o->s, s);
	return o;
}

obj_t *mkname(const char *s)
{
	obj_t *o = mkobj(T_NAME);
	o->s = GC_malloc_atomic(strlen(s) + 1);
	strcpy(o->s, s);
	return o;
}

obj_t *mkename(const char *s)
{
	obj_t *o = mkname(s);
	o->exec = 1;
	return o;
}

obj_t *mkop(op_t op)
{
	obj_t *o = mkobj(T_OP);
	o->op = op;
	return o;
}

obj_t *mkgstate(gstate_t *g)
{
	obj_t *o = mkobj(T_GSTATE);
	o->gs = GC_malloc(sizeof *g);
	memcpy(o->gs, g, sizeof *g);
	o->gs->path = path_copy(g->path);
	o->gs->clippath = path_copy(g->clippath);
	return o;
}

static inline objcmp_t icmp(int a, int b)
{
	if(a < b) return C_LT;
	if(a > b) return C_GT;
	return C_EQ;
}

static inline objcmp_t rcmp(double a, double b)
{
	if(a < b) return C_LT;
	if(a > b) return C_GT;
	return C_EQ;
}

static inline objcmp_t scmp(const char *a, const char *b)
{
	int v = strcmp(a, b);
	if(v < 0) return C_LT;
	if(v > 0) return C_GT;
	return C_EQ;
}

objcmp_t obj_cmp(obj_t *a, obj_t *b)
{
	if(a == b) return C_EQ;

	if(a->type == b->type) {
		switch(a->type) {
		case T_INT:
			return icmp(a->i, b->i);
		case T_REAL:
			return rcmp(a->r, b->r);
		case T_NAME:
		case T_STRING:
			return scmp(a->s, b->s);
		default:
			return C_INVALID;
		}
	}

	if((a->type == T_INT && b->type == T_REAL) || (a->type == T_REAL && b->type == T_INT)) {
		return rcmp(getreal(a), getreal(b));
	}

	if((a->type == T_STRING && b->type == T_NAME) || (a->type == T_NAME && b->type == T_STRING)) {
		return scmp(a->s, b->s);
	}

	return C_INVALID;
}

int obj_hash_eq(obj_t *a, obj_t *b)
{
	int i;

	if((a->type == T_STRING || a->type == T_NAME) && (b->type == T_STRING || b->type == T_NAME)) {
		return !strcmp(a->s, b->s);
	}

	if(a->type != b->type)
		return 0;

	switch(a->type) {
	case T_BOOL:
	case T_INT:
		return (a->i == b->i);
	case T_REAL:
		return (a->r == b->r);
	case T_NULL:
		return 1;
	case T_OP:
		return (a->op == b->op);
	case T_ARRAY:
		if(a->arr.size != b->arr.size) return 0;
		for(i=0; i<a->arr.size; i++)
			if(!obj_hash_eq(a->arr.v[i], b->arr.v[i])) return 0;
		return 1;
	default:
		return a == b;
	}
}

unsigned int obj_hash(obj_t *o)
{
	switch(o->type) {
	case T_NAME:
	case T_STRING:
		return string_hash(o->s);
	case T_BOOL:
	case T_INT:
		return MurmurHash2(&o->i, sizeof(o->i), 0);
	case T_REAL:
		return MurmurHash2(&o->r, sizeof(o->r), 0);
	case T_NULL:
		return 0;
	case T_OP:
		return MurmurHash2(&o->op, sizeof(o->op), 0);
	default:
		return MurmurHash2(o, sizeof(o), 0);
	}
	/* XXX: arrays ... */
}

obj_t * arr_idx(obj_t *arr, int i)
{
	assert(arr->type == T_ARRAY);
	assert(i >= 0 && i < arr->arr.size);
	return arr->arr.v[i];
}

void arr_setidx(obj_t *arr, int i, obj_t *o)
{
	assert(arr->type == T_ARRAY);
	assert(i >= 0 && i < arr->arr.size);
	arr->arr.v[i] = o;
}

obj_t * arr_peek(obj_t *arr)
{
	assert(arr->type == T_ARRAY);
	assert(arr->arr.size > 0);
	return arr->arr.v[arr->arr.size - 1];
}

void arr_add(obj_t *arr, obj_t *o)
{
	assert(arr->type == T_ARRAY);
	arr->arr.v = GC_realloc(arr->arr.v, (arr->arr.size + 1) * sizeof (*arr->arr.v));
	arr->arr.v[arr->arr.size] = o;
	arr->arr.size ++;
}

obj_t * arr_pop(obj_t *arr)
{
	assert(arr->type == T_ARRAY);
	assert(arr->arr.size > 0);

	obj_t *o = arr->arr.v[-- arr->arr.size];
	arr->arr.v = GC_realloc(arr->arr.v, arr->arr.size * sizeof (* arr->arr.v));
	return o;
}

obj_t * dict_lookup(obj_t *dict, obj_t *key)
{
	assert(dict->type = T_DICT);

	ptrhash_entry_t *entry = ptrhash_getentry(dict->dict, key);
	if(entry) return entry->value; else return NULL;
}

void dict_set(obj_t *dict, obj_t *key, obj_t *val)
{
	assert(dict->type == T_DICT);

	ptrhash_set(dict->dict, key, val);
}

spath_t * spath_init()
{
	spath_t *p = GC_malloc(sizeof *p);
	p->data = NULL;
	p->size = 0;
	p->cnt = 0;
	return p;
}

void spath_add(spath_t *p, double x, double y)
{
	if(p->size < ((p->cnt+1)*2)) {
		if(!p->size) {
			p->size = 16;
			p->data = GC_malloc_atomic(p->size * sizeof(double));
		} else {
			p->size *= 2;
			p->data = GC_realloc(p->data, p->size * sizeof(double));
		}
	}

	p->data[p->cnt*2] = x;
	p->data[p->cnt*2+1] = y;
	p->cnt ++;
}

spath_t * spath_copy(spath_t *p)
{
	spath_t *q = GC_malloc(sizeof *q);
	q->data = GC_malloc_atomic(q->size * sizeof(double));
	memcpy(q->data, p->data, p->cnt * 2 * sizeof(double));
	q->cnt = p->cnt;
	return q;
}

double *spath_pt(spath_t *p, int i)
{
	assert(i >= 0 && i < p->cnt);
	return p->data + (i * 2);
}

double *spath_last(spath_t *p)
{
	return spath_pt(p, p->cnt-1);
}

path_t * path_init()
{
	path_t *p = GC_malloc(sizeof *p);
	p->sp = NULL;
	p->cnt = 0;
	p->wasmoveto = 0;
	p->closed = 0;
	return p;
}

path_t * path_copy(path_t *p)
{
	path_t *q = GC_malloc(sizeof *p);
	q->cnt = p->cnt;
	q->sp = GC_malloc(q->cnt * sizeof *q->sp);
	q->wasmoveto = p->wasmoveto;
	q->closed = p->closed;
	int i;
	for(i=0; i<q->cnt; i++)
		q->sp[i] = spath_copy(p->sp[i]);
	return q;
}

spath_t * path_last(path_t *p)
{
	if(!p->cnt) return NULL;
	return p->sp[p->cnt-1];
}

void path_newsub(path_t *p)
{
	p->cnt ++;
	p->sp = GC_realloc(p->sp, p->cnt * sizeof *p->sp);
	p->sp[p->cnt-1] = spath_init();
	p->closed = 0;
}
