#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <math.h>
#include <sys/time.h>

#include "SDL_image.h"

#include "cmd.h"
#include "gfx.h"
#include "rgb.h"
#include "draw.h"
#include "util.h"
#include "area.h"

ptrhash_t *imagehash = NULL;
Uint32 colormap[256];

/* initialize image system */
void image_init()
{
	imagehash = strhash_new();

	int i;
	for(i=0;i<256; i++) colormap[i] = i | (i << 8) | (i << 16);
}

/* load image from "path", store it as "symbol" in the image hash */
void image_load(char *symbol, char *path)
{
	image_t *img;

	/* check if there's such an image already loaded */
	ptrhash_entry_t *entry = ptrhash_getentry(imagehash, symbol);
	if(entry) {
		img = entry->value;
		if(img->native != img->b32)
			SDL_FreeSurface(img->native);
		SDL_FreeSurface(img->b32);
		free(img);
	}

	/* load image */
	SDL_Surface *im = IMG_Load(path);
	if(!im) {
		fprintf(stderr, "Failed to load image %s from file %s\n%s\n", symbol, path, IMG_GetError());
		exit(1);
	}

	img = calloc(1, sizeof *img);
	if(!img) abort();

	img->alpha = im->flags & (SDL_SRCALPHA | SDL_SRCCOLORKEY);

	/* convert to 32 bpp pixel format for scaling */
	SDL_PixelFormat fmt = (SDL_PixelFormat) { .BitsPerPixel=32, .BytesPerPixel=4, .Rloss=0, .Gloss=0, .Bloss=0, .Aloss=0, .Rshift=16, .Gshift=8, .Bshift=0, .Ashift=24, .Rmask=0x00ff0000, .Gmask=0x0000ff00, .Bmask=0x000000ff, .Amask=0xff000000, .alpha = 0 };

	if(!img->alpha)
		fmt.Amask = 0;

	img->b32 = SDL_ConvertSurface(im, &fmt, img->alpha ? SDL_SRCALPHA : 0);
	if(!img->b32) {
		fprintf(stderr, "Failed to convert image %s (from file %s) to bpp32\n%s\n", symbol, path, SDL_GetError());
		exit(1);
	}

	/* convert to native format for blitting */
	if(img->alpha) {
		img->native = SDL_DisplayFormatAlpha(im);
		if(!img->native) {
			if(headless) {
				img->native = img->b32;
			} else {
				fprintf(stderr, "Failed to convert image %s (from file %s) to native format using DisplayFormatAlpha\n%s\n", symbol, path, SDL_GetError());
				exit(1);
			}
		}
	} else {
		img->native = SDL_DisplayFormat(im);
		if(!img->native) {
			if(headless) {
				img->native = img->b32;
			} else {
				fprintf(stderr, "Failed to convert image %s (from file %s) to native format using DisplayFormat\n%s\n", symbol, path, SDL_GetError());
				exit(1);
			}
		}
	}

	/*printf("loaded %s from %s, alpha=%d, natflags=%d, natdepth=%d, b32flags=%d\n", symbol, path, img->alpha, img->native->flags & (SDL_SRCALPHA | SDL_SRCCOLORKEY), img->native->format->BitsPerPixel, img->b32->flags & (SDL_SRCALPHA | SDL_SRCCOLORKEY));*/

	SDL_FreeSurface(im);

	/* set image in imagehash */
	if(entry) {
		entry->value = img;
	} else {
		ptrhash_set(imagehash, strdup(symbol), img);
	}
}

/* creates a new empty image */
void image_create(char *symbol, int w, int h)
{
	image_t *img;

	if(w < 1 || h < 1) {
		fprintf(stderr, "Tried to create image with width or height < 1\n");
		exit(1);
	}

	/* check if there's such an image already loaded */
	ptrhash_entry_t *entry = ptrhash_getentry(imagehash, symbol);
	if(entry) {
		img = entry->value;
		if(img->native != img->b32)
			SDL_FreeSurface(img->native);
		SDL_FreeSurface(img->b32);
		free(img);
	}

	img = calloc(1, sizeof *img);
	if(!img) abort();

	img->alpha = 0;

	img->b32 = SDL_CreateRGBSurface(SDL_SWSURFACE, w, h, 32, 0x00ff0000, 0x0000ff00, 0x000000ff, 0);

	SDL_FillRect(img->b32, NULL, background);

	img->native = SDL_CreateRGBSurface(SDL_SWSURFACE, w, h, main_surface->format->BitsPerPixel, main_surface->format->Rmask, main_surface->format->Gmask, main_surface->format->Bmask, main_surface->format->Amask);

	gfxclr bg = mkclr(background);
	SDL_FillRect(img->native, NULL, CLR(bg));

	/* set image in imagehash */
	if(entry) {
		entry->value = img;
	} else {
		ptrhash_set(imagehash, strdup(symbol), img);
	}
}

/* updates colormap */
void set_colormap(int index, Uint32 color)
{
	colormap[index] = color;
}

/* writes into image using colormap */
void image_update(char *symbol, int x, int y, char *string)
{
	ptrhash_entry_t *imge = ptrhash_getentry(imagehash, symbol);
	if(!imge) {
		fprintf(stderr, "image_set cannot find image: %s\n", symbol);
		exit(1);
	}
	image_t *img = imge->value;

	if(y < 0 || y >= img->native->h) return;

	int l = strlen(string);
	if(x < 1 - l) return;
	if(x >= img->native->w) return;

	char *p = string;
	if(x < 0) {
		p += (-x);
		x = 0;
	}

	void *p_b32 = (void*)img->b32->pixels + (y * img->b32->pitch + x * img->b32->format->BytesPerPixel);
	void *p_native = (void*)img->native->pixels + (y * img->native->pitch + x * img->native->format->BytesPerPixel);
	int native_bpp = img->native->format->BytesPerPixel;

	while(*p) {
		if(x == img->native->w) break;

		Uint32 color = colormap[(int)*p];
		int r = color >> 16;
		int g = (color >> 8) & 0xff;
		int b = color & 0xff;

		*(Uint32*)p_b32 = SDL_MapRGB(img->b32->format, r, g, b);
		p_b32 += 4;

		if(img->native != img->b32) {
			Uint32 ncolor = SDL_MapRGB(img->native->format, r, g, b);

			switch(native_bpp) {
			case 1:
				*(Uint8*)p_native = ncolor;
				p_native += 1;
				break;
			case 2:
				*(Uint16*)p_native = ncolor;
				p_native += 2;
				break;
			case 3:
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
				*(Uint8*)p_native = (ncolor >> 16) & 0xff;
				*(Uint8*)(p_native+1) = (ncolor >> 8) & 0xff;
				*(Uint8*)(p_native+2) = ncolor & 0xff;
				p_native += 3;
#else
				*(Uint8*)(p_native+2) = (ncolor >> 16) & 0xff;
				*(Uint8*)(p_native+1) = (ncolor >> 8) & 0xff;
				*(Uint8*)p_native = ncolor & 0xff;
				p_native += 3;
#endif
				break;
			case 4:
				*(Uint32*)p_native = ncolor;
				p_native += 4;
				break;
			}
		}

		p++;
		x++;
	}
}

matrixstack_t *matrixstack = NULL;

void transform(double x, double y, double *ox, double *oy)
{
	assert(matrixstack);
	double *m = matrixstack->m.v;
	*ox = x * m[0] + y * m[1] + m[2];
	*oy = x * m[3] + y * m[4] + m[5];
}

int color_lookup(Uint32 *out, char *name)
{
	int color = rgblookup(name);
	if(color < 0) {
		yyerror("named color '%s' not found", name);
		return 1;
	}

	namedcolor_t *c = namedcolors + color;
	*out = (c->r << 16) + (c->g << 8) + c->b;

	return 0;
}

int color_hex(Uint32 *out, char *hex)
{
	hex ++;
	int l = strlen(hex);
	Uint32 clr, x;
	int r, g, b;
	switch(l) {
	case 6:
		clr = strtol(hex, NULL, 16);
		break;
	case 3:
		x = strtol(hex, NULL, 16);
		r = (x&15);
		g = (x&0xf0) >> 4;
		b = (x >> 8)&15;
		clr = r|(r<<4)|(g<<8)|(g<<12)|(b<<16)|(b<<20);
		break;
	case 2:
		x = strtol(hex, NULL, 16);
		clr = x | (x << 8) | (x << 16);
		break;
	case 1:
		x = strtol(hex, NULL, 16);
		clr = x|(x<<4)|(x<<8)|(x<<12)|(x<<16)|(x<<20);
		break;
	default:
		abort();
	}
	*out = clr;

	return 0;
}

void set_color(Uint32 clr)
{
	drawcmd_t c;
	c.type = DCOLOR;
	c.color = mkclr(clr);
	draw(&c);
}

void set_alpha(Uint8 alpha)
{
	drawcmd_t c;
	c.type = DALPHA;
	c.alpha = alpha;
	draw(&c);
}

void set_dash(unsigned int dash)
{
	drawcmd_t c;
	c.type = DDASH;
	c.dash = dash;
	draw(&c);
}

void set_bg(Uint32 clr)
{
	background = clr;
}

void set_thick(float thick)
{
	drawcmd_t c;
	c.type = DTHICK;
	c.thick = thick;
	draw(&c);
}

void pushmatrix()
{
	matrixstack_t *s = calloc(1, sizeof *s);
	s->next = matrixstack;

	if(matrixstack) {
		s->m = matrixstack->m;
	} else {
		s->m = (matrix_t) {
			.v = { 1.0, 0, 0,
			       0, 1.0, 0,
			       0, 0, 1.0 }
		};
	}

	matrixstack = s;
}

void popmatrix()
{
	assert(matrixstack);
	if(!matrixstack->next) {
		fprintf(stderr, "Can't pop last matrix\n");
		exit(1);
	}

	matrixstack_t *c = matrixstack;
	matrixstack = matrixstack->next;
	free(c);
}

void multmatrix(matrix_t *m)
{
	assert(matrixstack);
	double *am = matrixstack->m.v;
	double *bm = m->v;
	double rm[9];

	rm[0] = am[0] * bm[0] + am[1] * bm[3] + am[2] * bm[6];
	rm[1] = am[0] * bm[1] + am[1] * bm[4] + am[2] * bm[7];
	rm[2] = am[0] * bm[2] + am[1] * bm[5] + am[2] * bm[8];

	rm[3] = am[3] * bm[0] + am[4] * bm[3] + am[5] * bm[6];
	rm[4] = am[3] * bm[1] + am[4] * bm[4] + am[5] * bm[7];
	rm[5] = am[3] * bm[2] + am[4] * bm[5] + am[5] * bm[8];

	rm[6] = am[6] * bm[0] + am[7] * bm[3] + am[8] * bm[6];
	rm[7] = am[6] * bm[1] + am[7] * bm[4] + am[8] * bm[7];
	rm[8] = am[6] * bm[2] + am[7] * bm[5] + am[8] * bm[8];

	memcpy((void *) am, (void *) rm, sizeof(rm));
}

#define PI_180 .0174532925199432957692369

void rotate(double angle)
{
	int q;
	double a = remquo(angle, 90.0, &q) * PI_180;
	double ca = cos(a);
	double sa = sin(a);
	matrix_t m;

	switch (q&3) {
	case 0:
		m = (matrix_t){ .v = {
			ca, -sa, 0,
			sa, ca, 0,
			0, 0, 1.0
		}};
		break;
	case 1:
		m = (matrix_t){ .v = {
			-sa, -ca, 0,
			ca, -sa, 0,
			0, 0, 1.0
		}};
		break;
	case 2:
		m = (matrix_t){ .v = {
			-ca, sa, 0,
			-sa, -ca, 0,
			0, 0, 1.0
		}};
		break;
	case 3:
		m = (matrix_t){ .v = {
			sa, ca, 0,
			-ca, sa, 0,
			0, 0, 1.0
		}};
		break;
	}

	multmatrix(&m);
}


void scale(double x, double y)
{
	matrix_t m = (matrix_t) {
		.v = {
			x, 0, 0,
			0, y, 0,
			0, 0, 1.0
		}
	};
	multmatrix(&m);
}

void shift(double x, double y)
{
	matrix_t m = (matrix_t) {
		.v = {
			1.0, 0, x,
			0, 1.0, y,
			0, 0, 1.0
		}
	};
	multmatrix(&m);
}

void viewport(double left, double bottom, double width, double height)
{
	/* shift -1 -1 * scale 2/width 2/height * shift -left -bottom */
	double w = 2.0/width;
	double h = 2.0/height;
	matrix_t m = (matrix_t) {
		.v = {
			w, 0, -1.0 - left*w,
			0, h, -1.0 - bottom*h,
			0, 0, 1
		}
	};

	multmatrix(&m);
}

void viewport_macro(char *name, int aspect)
{
	matrix_t m;
	macro_boundingbox(name, aspect, &m);
	multmatrix(&m);
}

void loadident()
{
	assert(matrixstack);
	matrixstack->m = (matrix_t) { .v = { 1.0, 0, 0,
	                                     0, 1.0, 0,
	                                     0, 0, 1.0 } };
}

void eventreport(int on)
{
	eventreporting = on;
	if(on) {
		assert(matrixstack);
		eventmatrix = matrixstack->m;
	}
}

void frame()
{
	beginframe();
}

void flush()
{
	flushframe(1);
	if(eventreporting) fflush(stdout);
}

void drawdot(double x, double y)
{
	drawcmd_t c;
	c.type = DDOT;
	transform(x, y, &c.dot.x, &c.dot.y);
	draw(&c);
}

void drawline(double x1, double y1, double x2, double y2)
{
	drawcmd_t c;
	c.type = DLINE;
	transform(x1, y1, &c.line.x1, &c.line.y1);
	transform(x2, y2, &c.line.x2, &c.line.y2);
	draw(&c);
}

void drawpoly(polygon_t *p)
{
	drawcmd_t c;
	c.type = DPOLY;
	c.poly = p;

	int i, j;
	for(i=0, j=0; i<p->npoints; i++, j+=2) {
		double x, y;
		transform(p->points[j], p->points[j+1], &x, &y);
		p->points[j] = x;
		p->points[j+1] = y;
	}

	draw(&c);
}

static double linesegx, linesegy;
void drawlinefrom(double x, double y)
{
	linesegx = x;
	linesegy = y;
}

void drawlineto(double x, double y)
{
	drawline(linesegx, linesegy, x, y);
	linesegx = x;
	linesegy = y;
}

void drawarrow(double x1, double y1, double x2, double y2)
{
	double nx1, ny1, nx2, ny2;

	/* arrowhead dimensions are best done in screen coordinates
	 * so the arrowhead stays the same size ... */

	drawcmd_t c;
	c.type = DLINE;

	transform(x1, y1, &nx1, &ny1);
	transform(x2, y2, &nx2, &ny2);

	c.line.x1 = nx1;
	c.line.y1 = ny1;
	c.line.x2 = nx2;
	c.line.y2 = ny2;
	draw(&c);

	double vx = nx2 - nx1;
	double vy = ny2 - ny1;

	double vsize = hypot(vx, vy);

	if(vsize > 0) {
		double nx = -vy;
		double ny = vx;

		double head = 0.02 / vsize;

		double hvx = vx * head;
		double hvy = vy * head;
		double hnx = nx * head;
		double hny = ny * head;

		c.line.x1 = nx2; c.line.y1 = ny2; c.line.x2 = nx2+hnx-hvx; c.line.y2 = ny2+hny-hvy; draw(&c);
		c.line.x1 = nx2; c.line.y1 = ny2; c.line.x2 = nx2-hnx-hvx; c.line.y2 = ny2-hny-hvy; draw(&c);
	}
}

void drawrect(double cx, double cy, double w, double h)
{
	double x1 = cx;
	double y1 = cy;
	double x2 = cx + w;
	double y2 = cy + h;
	drawrect2(x1, y1, x2, y2);
}

void drawrect2(double x1, double y1, double x2, double y2)
{
	drawline(x1, y1, x2, y1);
	drawline(x2, y1, x2, y2);
	drawline(x2, y2, x1, y2);
	drawline(x1, y2, x1, y1);
}

void drawtext(double x, double y, int horalign, int veralign, const char *text)
{
	drawcmd_t c;
	c.type = DTEXT;
	c.text.horalign = horalign;
	c.text.veralign = veralign;
	c.text.str = strdup(text);
	transform(x, y, &c.text.x, &c.text.y);
	draw(&c);
}

void drawblit(double x, double y, int horalign, int veralign, char *image)
{
	drawcmd_t c;
	c.type = DBLIT;
	c.text.horalign = horalign;
	c.text.veralign = veralign;
	c.text.str = strdup(image);
	transform(x, y, &c.text.x, &c.text.y);
	draw(&c);
}

void drawimgpoly(double ax, double ay, double ux, double uy, double vx, double vy, char *img, polygon_t *p)
{
	drawcmd_t c;
	c.type = DIMGPOLY;

	ux += ax;
	uy += ay;
	vx += ax;
	vy += ay;

	transform(ax, ay, &c.imgpoly.ax, &c.imgpoly.ay);
	transform(ux, uy, &c.imgpoly.ux, &c.imgpoly.uy);
	transform(vx, vy, &c.imgpoly.vx, &c.imgpoly.vy);

	c.imgpoly.img = img;
	c.imgpoly.poly = p;

	int i, j;
	for(i=0, j=0; i<p->npoints; i++, j+=2) {
		double x, y;
		transform(p->points[j], p->points[j+1], &x, &y);
		p->points[j] = x;
		p->points[j+1] = y;
	}

	draw(&c);
}

void clip(double x, double y, double w, double h)
{
	drawcmd_t c;
	c.type = DCLIP;

	double x2 = x + w;
	double y2 = y + h;

	transform(x, y, &c.line.x1, &c.line.y1);
	transform(x2, y2, &c.line.x2, &c.line.y2);

	draw(&c);
}

/* parse error callback */
void yyerror(const char *str, ...)
{
	char buf[2048];
	va_list(ap);

	sprintf(buf, "Input error (line %d): ", yyline);
	char *b = buf + strlen(buf);

	va_start(ap, str);
	vsprintf(b, str, ap);
	va_end(ap);

	strcat(buf, "\n");

	fprintf(stderr, "%s", buf);

	exit(1);
}

void macro_execute(char *name)
{
	ptrhash_entry_t *entry = ptrhash_getentry(macrohash, name);
	if(!entry) {
		fprintf(stderr, "macro \"%s\" not found\n", name);
		exit(1);
	}

	macro_t *m = entry->value;

	int i;
	macro_el_t *el;
	for(i=0, el=m->els; i<m->cnt; i++, el++) {
		switch(el->type) {
		case MCOLOR:
			{
				drawcmd_t c;
				c.type = DCOLOR;
				c.color = el->color;
				draw(&c);
			}
			break;
		case MDASH:
			{
				drawcmd_t c;
				c.type = DDASH;
				c.dash = el->dash;
				draw(&c);
			}
			break;
		case MALPHA:
			{
				drawcmd_t c;
				c.type = DALPHA;
				c.alpha = el->alpha;
				draw(&c);
			}
			break;
		case MTHICK:
			{
				drawcmd_t c;
				c.type = DTHICK;
				c.thick = el->thick;
				draw(&c);
			}
			break;
		case MDOT:
			drawdot(el->dot.x, el->dot.y);
			break;
		case MLINE:
			drawline(el->line.x1, el->line.y1, el->line.x2, el->line.y2);
			break;
		case MARROW:
			drawarrow(el->line.x1, el->line.y1, el->line.x2, el->line.y2);
			break;
		case MTEXT:
			drawtext(el->text.x, el->text.y, el->text.horalign, el->text.veralign, el->text.str);
			break;
		case MBLIT:
			drawblit(el->text.x, el->text.y, el->text.horalign, el->text.veralign, el->text.str);
			break;
		case MIMGPOLY:
			drawimgpoly(el->imgpoly.ax, el->imgpoly.ay, el->imgpoly.ux, el->imgpoly.uy, el->imgpoly.vx, el->imgpoly.vy, strdup(el->imgpoly.img), polygon_copy(el->imgpoly.poly));
			break;
		case MPUSH:
			pushmatrix();
			break;
		case MPOP:
			popmatrix();
			break;
		case MMULT:
			multmatrix(&el->matrix);
			break;
		case MINVOKE:
			macro_execute(el->invoke);
			break;
		case MMACROVIEWPORT:
			viewport_macro(el->viewport_macro.name, el->viewport_macro.aspect);
			break;
		case MPOLY:
			drawpoly(polygon_copy(el->poly));
			break;
		case MIDENT:
			loadident();
			break;
		case MAREA:
			set_area(strdup(el->area.name), polygon_copy(el->area.poly));
			break;
		case MDELAREA:
			del_area(el->area.name);
			break;
		case MCLIP:
			clip(el->clip.x, el->clip.y, el->clip.w, el->clip.h);
			break;
		case MCOLORMAP:
			set_colormap(el->colormap.index, el->colormap.color);
			break;
		case MIMAGESET:
			image_update(el->image_set.img, el->image_set.x, el->image_set.y, el->image_set.string);
			break;
		default:;
		}
	}
}

/* writes a matrix into *out_matrix that keeps the named macro inside the
 * -1, -1 ... 1, 1 bounding box.
 * if aspect is nonzero, aspect ratio is kept */
void macro_boundingbox(char *name, int aspect, matrix_t *out_matrix)
{
	matrixstack_t *old_stack = matrixstack;

	matrixstack = calloc(1, sizeof *matrixstack);
	matrixstack->m = (matrix_t) {
		.v = { 1.0, 0, 0,
			   0, 1.0, 0,
			   0, 0, 1.0 }
	};

	double xmin = INFINITY;
	double xmax = -INFINITY;
	double ymin = INFINITY;
	double ymax = -INFINITY;

	macro_boundingbox_expand(name, &xmin, &ymin, &xmax, &ymax);

	while(matrixstack->next)
		popmatrix();
	free(matrixstack);
	matrixstack = old_stack;

	if(xmin == INFINITY) {
		*out_matrix = (matrix_t) {
			.v = { 1.0, 0, 0,
				   0, 1.0, 0,
				   0, 0, 1.0 }
		};
		return;
	}

	double width = xmax - xmin;
	double height = ymax - ymin;

	if(aspect) {
		if(width > height) {
			ymin += ((height - width) / 2);
			height = width;
		} else {
			xmin += ((width - height) / 2);
			width = height;
		}
	}

	double w = 2.0 / width;
	double h = 2.0 / height;
	*out_matrix = (matrix_t) {
		.v = {
			w, 0, -1.0 - xmin*w,
			0, h, -1.0 - ymin*h,
			0, 0, 1
		}
	};
}

static double * polybuffer = NULL;
static int polybuffer_cnt = 0;
static int polybuffer_alloc = 0;

void polybuffer_add(double x, double y)
{
	if(polybuffer_cnt == polybuffer_alloc) {
		if(!polybuffer_alloc) polybuffer_alloc = 16; else polybuffer_alloc *= 2;
		polybuffer = realloc(polybuffer, polybuffer_alloc * sizeof *polybuffer);
		if(!polybuffer) abort();
	}

	polybuffer[polybuffer_cnt ++] = x;
	polybuffer[polybuffer_cnt ++] = y;
}

polygon_t *polybuffer_end()
{
	polygon_t *polygon = malloc(sizeof(*polygon) + polybuffer_cnt * sizeof(double));
	if(!polygon) abort();
	polygon->aa = 0;
	polygon->npoints = polybuffer_cnt / 2;
	memcpy(polygon->points, polybuffer, polybuffer_cnt * sizeof (double));

	free(polybuffer);
	polybuffer = NULL;
	polybuffer_cnt = 0;
	polybuffer_alloc = 0;

	return polygon;
}

polygon_t *polygon_copy(polygon_t *p)
{
	size_t size = sizeof(polygon_t) + p->npoints * 2 * sizeof(double);
	polygon_t *q = malloc(size);
	if(!q) abort();
	memcpy(q, p, size);
	return q;
}

polygon_t *makerect(double cx, double cy, double w, double h)
{
	return makerect2(cx, cy, cx+w, cy+h);
}

polygon_t *makerect2(double x1, double y1, double x2, double y2)
{
	polygon_t *p = malloc(sizeof *p + 8 * sizeof(double));
	p->aa = 0;
	p->npoints = 4;
	p->points[0] = x1; p->points[1] = y1;
	p->points[2] = x2; p->points[3] = y1;
	p->points[4] = x2; p->points[5] = y2;
	p->points[6] = x1; p->points[7] = y2;

	return p;
}

polygon_t *makecircle(double cx, double cy, double r, double verts)
{
	int v = (int) verts;
	if(v < 3) v = 3;

	polygon_t *p = malloc(sizeof *p + v * 2 * sizeof(double));
	if(!p) abort();
	p->aa = 0;
	p->npoints = v;

	int i, j;
	double step = 3.14159265358979 * 2.0 / (double) v;
	double a = 0;
	for(i=0, j=0; i<v; i++) {
		p->points[j++] = cos(a) * r + cx;
		p->points[j++] = sin(a) * r + cy;
		a += step;
	}

	return p;
}

void drawcircle(double cx, double cy, double r, double verts)
{
	int v = (int) verts;
	if(v < 3) v = 3;
	double step = 3.14159265358979 * 2.0 / (double) v;
	double a = step;
	double fx = r + cx;
	double fy = cy;
	double ox = fx;
	double oy = fy;
	int i;
	for(i=1; i<v; i++) {
		double x = cos(a) * r + cx;
		double y = sin(a) * r + cy;
		drawline(ox, oy, x, y);
		ox = x;
		oy = y;
		a += step;
	}
	drawline(ox, oy, fx, fy);
}

/* returns time since epoch in seconds */
double dbltime(void)
{
	struct timeval tv;
	gettimeofday(&tv, NULL);
	return (double) tv.tv_sec + (double) tv.tv_usec / 1000000.0;
}

ptrhash_t *cursorhash = NULL;

/* initialize mouse cursors */
void cursor_init()
{
	cursorhash = strhash_new();
}

/* define a new mouse cursor */
void defcursor(char *name, const char *data, const char *mask, int hot_x, int hot_y)
{
	SDL_Cursor *cur;

	/* check if there's such a cursor already loaded */
	ptrhash_entry_t *entry = ptrhash_getentry(cursorhash, name);
	if(entry) {
		cur = entry->value;
		SDL_FreeCursor(cur);
	}

	if(strlen(data) != 256 || strlen(mask) != 256)
		yyerror("defcursor: both data and mask arguments have to be 256 characters long (32*32=1024 bits in hex), in cursor %s", name);

	Uint8 cdata[128];
	Uint8 cmask[128];

	int i;
	for(i=0; i<128; i++) {
		char sd[] = { data[i*2], data[i*2+1], 0 };
		char sm[] = { mask[i*2], mask[i*2+1], 0 };

		char *end;
		int v;

		v = strtol(sd, &end, 16);
		if(*end) yyerror("defcursor: failed to parse hex %s in data, in cursor %s", sd, name);
		cdata[i] = v;

		v = strtol(sm, &end, 16);
		if(*end) yyerror("defcursor: failed to parse hex %s in mask, in cursor %s", sm, name);
		cmask[i] = v;
	}

	cur = SDL_CreateCursor(cdata, cmask, 32, 32, hot_x, hot_y);
	if(!cur) yyerror("defcursor: SDL_CreateCursor failed");

	/* set cursor in hash */
	if(entry) {
		entry->value = cur;
	} else {
		ptrhash_set(cursorhash, strdup(name), cur);
	}
}

/* set current mouse cursor */
void setcursor(char *name)
{
	ptrhash_entry_t *entry = ptrhash_getentry(cursorhash, name);
	if(entry == NULL)
		yyerror("setcursor: cursor \"%s\" not found", name);

	SDL_SetCursor(entry->value);
}
