%{
	#include <stdlib.h>
	#include <stdio.h>

	#include "cmd.h"
	#include "area.h"

	int yylex();

	static Uint32 ncolor(double r, double g, double b)
	{
		inline int cb(float x) {
			int z = x * 255.0;
			if(z < 0) z = 0;
			else if(z > 255) z = 255;
			return z;
		}
		int ir = cb(r);
		int ig = cb(g);
		int ib = cb(b);
		return (ir << 16) | (ig << 8) | ib;
	}
%}

%union {
	double d;
	char *s;
	int i;
	Uint32 ui32;
	polygon_t *p;
}

%token <d> NUMBER
%token <i> HEX
%token <s> SYMBOL COLORHEX
%token <i> CHARACTER

%token LINE ARROW RECT TEXT DOT LINES POLY APOLY
%token CIRCLE FILLCIRCLE FILLRECT
%token LEFT CENTER RIGHT BOTTOM TOP
%token ALPHA COLOR COLORMAP BG THICK DASH
%token PUSH POP MULT ROTATE SCALE SHIFT IDENT VIEWPORT
%token EVENTS ASPECT ON OFF ECHOCMD ECHOPTCMD AREA DELAREA KILLAREAS
%token FRAME FLUSH FLUSHDELAY SCREENSHOT SCREENDUMP
%token SLEEP TITLE
%token KEEPALIVE QUIT RESIZE WARP
%token MACRO ENDMACRO FORGET INVOKE
%token IMAGE IMAGE_SET BLIT IMGPOLY IMGAPOLY
%token DEFCURSOR SETCURSOR
%token CLIP

%type <ui32> colordef
%type <i> horalign veralign
%type <p> polygon
%type <i> alphadef

%%

input:
	| input stmt
;

/* color definition */
colordef:
	SYMBOL      { if(color_lookup(&$$, $1)) YYABORT; }
	| COLORHEX  { if(color_hex(&$$, $1)) YYABORT; }
	| NUMBER    { $$=ncolor($1, $1, $1); }
	| NUMBER NUMBER NUMBER { $$=ncolor($1, $2, $3); }
;

/* alpha definition */
alphadef:
	NUMBER
	{
		$$ = $1 * 255.0;
		if($$ < 0) $$ = 0;
		if($$ > 255) $$ = 255;
	}
;

/* main statement */

stmt:
	  COLOR colordef { set_color($2); }
	| BG colordef    { set_bg($2); }
	| THICK NUMBER   { set_thick($2); }
	| ALPHA alphadef { set_alpha($2); }
	| DASH HEX { set_dash($2); }

	| PUSH { pushmatrix(); }
	| POP  { popmatrix(); }

	| MULT NUMBER NUMBER NUMBER
	       NUMBER NUMBER NUMBER
	       NUMBER NUMBER NUMBER
	{ multmatrix(& (matrix_t) {
	     .v = { $2, $3, $4,
	            $5, $6, $7,
	            $8, $9, $10 } }); }

	| ROTATE NUMBER        { rotate($2); }
	| SCALE NUMBER         { scale($2, $2); }
	| SCALE NUMBER NUMBER  { scale($2, $3); }
	| SHIFT NUMBER NUMBER  { shift($2, $3); }
	| IDENT                { loadident(); }
	| VIEWPORT NUMBER NUMBER NUMBER NUMBER { viewport($2, $3, $4, $5); }
	| VIEWPORT NUMBER NUMBER '-' NUMBER NUMBER { viewport($2, $3, $5-$2, $6-$3); }
	| VIEWPORT SYMBOL { viewport_macro($2, 0); }
	| VIEWPORT ASPECT SYMBOL { viewport_macro($3, 1); }
	| CLIP NUMBER NUMBER NUMBER NUMBER { clip($2, $3, $4, $5); }
	| CLIP NUMBER NUMBER '-' NUMBER NUMBER { clip($2, $3, $5-$2, $6-$3); }

	| EVENTS ON  { eventreport(1); }
	| EVENTS OFF { eventreport(0); }

	| ASPECT ON  { keepaspect = 1; }
	| ASPECT OFF { keepaspect = 0; }

	| KEEPALIVE ON  { keepalive = 1; }
	| KEEPALIVE OFF { keepalive = 0; }

	| QUIT { exit(0); }

	| ECHOCMD SYMBOL { printf("%s\n", $2); fflush(stdout); }
	| ECHOPTCMD SYMBOL { printf("%s", $2); } echopts '.' { printf("\n"); fflush(stdout); }

	| TITLE SYMBOL { SDL_WM_SetCaption($2, $2); }

	| RESIZE NUMBER NUMBER { resize($2, $3); }
	| WARP NUMBER NUMBER { warpmouse($2, $3); }
	| WARP AREA SYMBOL { if(warp_area($3)) YYABORT; }

	| FRAME { frame(); }
	| FLUSHDELAY NUMBER { flushdelay($2); }
	| FLUSH { flush(); }
	| SCREENSHOT SYMBOL { saveframeas($2); } 
	| SCREENDUMP '-' { dumpframe(NULL); }
	| SCREENDUMP SYMBOL { dumpframe($2); }

	| DOT NUMBER NUMBER
	{ drawdot($2, $3); }
	| LINE NUMBER NUMBER NUMBER NUMBER
	{ drawline($2, $3, $4, $5); }
	| ARROW NUMBER NUMBER NUMBER NUMBER
	{ drawarrow($2, $3, $4, $5); }
	| RECT NUMBER NUMBER NUMBER NUMBER
	{ drawrect($2, $3, $4, $5); }
	| RECT NUMBER NUMBER '-' NUMBER NUMBER
	{ drawrect2($2, $3, $5, $6); }
	| TEXT NUMBER NUMBER SYMBOL
	{ drawtext($2, $3, 0, 0, $4); }
	| TEXT horalign veralign NUMBER NUMBER SYMBOL
	{ drawtext($4, $5, $2, $3, $6); }
	| LINES NUMBER NUMBER { drawlinefrom($2, $3); } linesegs

	| SLEEP NUMBER
	{ SDL_Delay($2 * 1000); }

	| MACRO SYMBOL ENDMACRO
	{ begin_macro($2, 0); end_macro(0); }

	| MACRO SYMBOL
	{ begin_macro($2, 0); }
	  macrobody
	  ENDMACRO
	{ end_macro(0); }

	| MACRO PUSH SYMBOL ENDMACRO
	{ begin_macro($3, 1); end_macro(1); }

	| MACRO PUSH SYMBOL
	{ begin_macro($3, 1); }
	  macrobody
	  ENDMACRO
	{ end_macro(1); }

	| FORGET
	{ kill_macros(); }

	| INVOKE SYMBOL
	{ macro_execute($2); }

	| INVOKE NUMBER NUMBER SYMBOL
	{ pushmatrix();
	  shift($2, $3);
	  macro_execute($4);
	  popmatrix();
	}

	| POLY polygon
	{ drawpoly($2); }
	| APOLY polygon
	{ $2->aa = 1; drawpoly($2); }

	| IMGPOLY SYMBOL { $2 = strdup($2); } polygon
	{
		drawimgpoly($4->points[0], $4->points[1], 1, 0, 0, 1, $2, $4);
	}

	| IMGPOLY NUMBER NUMBER SYMBOL { $4 = strdup($4); } polygon
	{
		drawimgpoly($6->points[0], $6->points[1], $2, 0, 0, $3, $4, $6);
	}

	| IMGPOLY NUMBER NUMBER NUMBER NUMBER SYMBOL { $6 = strdup($6); } polygon
	{
		drawimgpoly($8->points[0], $8->points[1], $2, $3, $4, $5, $6, $8);
	}

	| IMGPOLY NUMBER NUMBER NUMBER NUMBER NUMBER NUMBER SYMBOL { $8 = strdup($8); } polygon
	{
		drawimgpoly($2, $3, $4, $5, $6, $7, $8, $10);
	}

	| FILLRECT NUMBER NUMBER NUMBER NUMBER
	{ drawpoly(makerect($2, $3, $4, $5)); }
	| FILLRECT NUMBER NUMBER '-' NUMBER NUMBER
	{ drawpoly(makerect2($2, $3, $5, $6)); }
	| CIRCLE NUMBER NUMBER NUMBER NUMBER
	{ drawcircle($2, $3, $4, $5); }
	| FILLCIRCLE NUMBER NUMBER NUMBER NUMBER
	{ drawpoly(makecircle($2, $3, $4, $5)); }

	| AREA SYMBOL { $2 = strdup($2); } polygon
	{ set_area($2, $4); }

	| DELAREA SYMBOL
	{ del_area($2); }

	| KILLAREAS
	{ kill_areas(); }

	| IMAGE SYMBOL { $2 = strdup($2); } SYMBOL
	{ image_load($2, $4); free($2); }

	| IMAGE NUMBER NUMBER SYMBOL
	{ image_create($4, $2, $3); }

	| IMAGE_SET SYMBOL { $2 = strdup($2); } NUMBER NUMBER SYMBOL
	{ image_update($2, $4, $5, $6); free($2); }

	| COLORMAP CHARACTER colordef
	{ set_colormap($2, $3); }

	| BLIT NUMBER NUMBER SYMBOL
	{ drawblit($2, $3, 0, 0, $4); }

	| BLIT horalign veralign NUMBER NUMBER SYMBOL
	{ drawblit($4, $5, $2, $3, $6); }

	| DEFCURSOR
		SYMBOL { $2 = strdup($2); }
		SYMBOL { $4 = strdup($4); }
		SYMBOL
		NUMBER NUMBER
	{
		defcursor($2, $4, $6, $7, $8);
		free($2);
		free($4);
	}

	| SETCURSOR SYMBOL { setcursor($2); }
;

linesegs:
	| linesegs lineseg
;

lineseg:
	NUMBER NUMBER { drawlineto($1, $2); }
;

macrobody:
	macro
	| macrobody macro

macro:
	  COLOR colordef
	{ macro_color($2); }
	| THICK NUMBER
	{ macro_set_thick($2); }
	| ALPHA alphadef
	{ macro_alpha($2); }
	| DASH HEX
	{ macro_dash($2); }
	| DOT NUMBER NUMBER
	{ macro_dot($2, $3); }
	| LINE NUMBER NUMBER NUMBER NUMBER
	{ macro_line($2, $3, $4, $5); }
	| ARROW NUMBER NUMBER NUMBER NUMBER
	{ macro_arrow($2, $3, $4, $5); }
	| RECT NUMBER NUMBER NUMBER NUMBER
	{ macro_rect($2, $3, $4, $5); }
	| RECT NUMBER NUMBER '-' NUMBER NUMBER
	{ macro_rect2($2, $3, $5, $6); }
	| TEXT NUMBER NUMBER SYMBOL
	{ macro_text($2, $3, 0, 0, $4); }
	| TEXT horalign veralign NUMBER NUMBER SYMBOL
	{ macro_text($4, $5, $2, $3, $6); }
	| LINES NUMBER NUMBER { macrolinefrom($2, $3); } macrolinesegs

	| BLIT NUMBER NUMBER SYMBOL
	{ macro_blit($2, $3, 0, 0, $4); }

	| BLIT horalign veralign NUMBER NUMBER SYMBOL
	{ macro_blit($4, $5, $2, $3, $6); }

	| IMGPOLY SYMBOL { $2 = strdup($2); } polygon
	{
		macro_imgpoly($4->points[0], $4->points[1], 1, 0, 0, 1, $2, $4);
	}

	| IMGPOLY NUMBER NUMBER SYMBOL { $4 = strdup($4); } polygon
	{
		macro_imgpoly($6->points[0], $6->points[1], $2, 0, 0, $3, $4, $6);
	}

	| IMGPOLY NUMBER NUMBER NUMBER NUMBER SYMBOL { $6 = strdup($6); } polygon
	{
		macro_imgpoly($8->points[0], $8->points[1], $2, $3, $4, $5, $6, $8);
	}

	| IMGPOLY NUMBER NUMBER NUMBER NUMBER NUMBER NUMBER SYMBOL { $8 = strdup($8); } polygon
	{
		macro_imgpoly($2, $3, $4, $5, $6, $7, $8, $10);
	}

	| PUSH
	{ macro_push(); }
	| POP
	{ macro_pop(); }

	| MULT NUMBER NUMBER NUMBER
	       NUMBER NUMBER NUMBER
	       NUMBER NUMBER NUMBER
	{ macro_mult(& (matrix_t) {
	     .v = { $2, $3, $4,
	            $5, $6, $7,
	            $8, $9, $10 } }); }

	| ROTATE NUMBER        { macro_rotate($2); }
	| SCALE NUMBER         { macro_scale($2, $2); }
	| SCALE NUMBER NUMBER  { macro_scale($2, $3); }
	| SHIFT NUMBER NUMBER  { macro_shift($2, $3); }
	| IDENT                { macro_loadident(); }
	| VIEWPORT NUMBER NUMBER NUMBER NUMBER { macro_viewport($2, $3, $4, $5); }
	| VIEWPORT NUMBER NUMBER '-' NUMBER NUMBER { macro_viewport($2, $3, $5-$2, $6-$3); }
	| VIEWPORT SYMBOL { macro_viewport_macro($2, 0); }
	| VIEWPORT ASPECT SYMBOL { macro_viewport_macro($3, 1); }
	| CLIP NUMBER NUMBER NUMBER NUMBER { macro_clip($2, $3, $4, $5); }
	| CLIP NUMBER NUMBER '-' NUMBER NUMBER { macro_clip($2, $3, $5-$2, $6-$3); }

	| INVOKE SYMBOL
	{ macro_invokemacro($2); }

	| INVOKE NUMBER NUMBER SYMBOL
	{ macro_push();
	  macro_shift($2, $3);
	  macro_invokemacro($4);
	  macro_pop();
	}

	| POLY polygon
	{ macro_poly($2); }
	| APOLY polygon
	{ $2->aa = 1; macro_poly($2); }
	| FILLRECT NUMBER NUMBER NUMBER NUMBER
	{ macro_poly(makerect($2, $3, $4, $5)); }
	| FILLRECT NUMBER NUMBER '-' NUMBER NUMBER
	{ macro_poly(makerect2($2, $3, $5, $6)); }
	| CIRCLE NUMBER NUMBER NUMBER NUMBER
	{ macro_circle($2, $3, $4, $5); }
	| FILLCIRCLE NUMBER NUMBER NUMBER NUMBER
	{ macro_poly(makecircle($2, $3, $4, $5)); }

	| AREA SYMBOL { $2 = strdup($2); } polygon
	{ macro_area($2, $4); }
	| DELAREA SYMBOL
	{ macro_delarea(strdup($2)); }

	| COLORMAP CHARACTER colordef
	{ macro_colormap($2, $3); }

	| IMAGE_SET SYMBOL { $2 = strdup($2); } NUMBER NUMBER SYMBOL
	{ macro_image_update($2, $4, $5, strdup($6)); }
;

macrolinesegs:
	| macrolinesegs macrolineseg
;

macrolineseg:
	NUMBER NUMBER { macrolineto($1, $2); }
;

echopts:
	| echopts echopt
;

echopt:
	NUMBER NUMBER
	{
		int ox, oy;
		transform_pixel($1, $2, &ox, &oy);
		printf(" %d %d", ox, oy);
	}
;

horalign:
	  LEFT   { $$ = -1; }
	| CENTER { $$ = 0; }
	| RIGHT  { $$ = 1; }
;

veralign:
	  BOTTOM { $$ = -1; }
	| CENTER { $$ = 0; }
	| TOP    { $$ = 1; }
;

polygon:
	polypoint polypoint polypoint polypoints
	{ $$ = polybuffer_end(); }
;

polypoints:
	| polypoints polypoint
;

polypoint:
	NUMBER NUMBER
	{ polybuffer_add($1, $2); }
;

%%
