#!/bin/sh

# Copyright (C) 2016, Tibor 'Igor2' Palinkas
# Distributed under the GNU GPL version 2.0 (or any later version)

help()
{
cat <<HELP_END
animpie - generate pie chart animator scripts
Animpie reads stdin for a pie script and dumps an animator script on stdout.

Usage: animpie [-e]
  -e         embed: do not generate frame and flush
  -h         this help screen
  -E         print an example pie script

Script syntax
~~~~~~~~~~~~~
Lines starting with # are comments, empty lines are ignored.
The rest of the input consists of pairs of lines. The first line of the pair
starts with a @ and is a command, the second line is the argument. A command
either sets a global (marked with G) state or a slice state (marged with S).
Command "@slice" creates a new slice.

Command table:

name     scope     argument   description

@slice    S        float,*    create a new slice of the given size 
@label    S        text       draw a line and print label text next to the slice
@color    S        color      slice fill color [animator syntax]
@pull     S        float      pull the slice out from the pie by animator coord
@total    G        float      total amount in the pie

@cx       G        float      center of the pie, X [animator coord]
@cy       G        float      center of the pie, X [animator coord]
@radius   G        float      radius of the pie [animator coord]
@minres   G        float      minimum resolution for arc approximation [radian]
@cthick   G        float      contour thickenss [animator syntax, 0=off]
@lthick   G        float      label line thickenss [animator syntax, 0=off]
@labarm   G        float      horizontal label line arm length [animator coord]
@labspc   G        float      label text vs. label line spacing [animator coord]

Slice sizes, total amount
~~~~~~~~~~~~~~~~~~~~~~~~~
If each slice has an explicit size, pie total is automatically calculated
and @total should be omitted. If @total is specified, it is ignored or
an error is generated if it differs from the calculated sum by more than 1%.

There may be one or more slices with size "*". In this case @total is
mandatory. Animpie will calculate the sum of explicit slice sizes and
what is left to @total is equally divided up among all the "*" sized
slices.

HELP_END
}

example()
{
cat <<EXAMPLE_END
@total
100
@slice
24
@label
foo
@color
red

@slice
47
@label
bar
@color
green

@slice
1
@label
small1
@color
grey

@slice
1
@label
small2
@color
grey
@pull
0.2

@slice
1
@label
small3
@color
grey

@slice
*
@label
baz1
@color
#001188
@slice
*
@label
baz2
@color
#002299
EXAMPLE_END
}

### MAIN ###

while test $# -gt 0
do
	case "$1" in
		-e) embed=1;;
		-E) example; exit 0;;
		-h|--help) help; exit 0;;
		*) echo "Unknown argument $1" >&2; exit 1;;
	esac
	shift 1
done

awk -v "embed=$embed" '
	BEGIN {
		pi = 3.141592654
		radius = 0.5
		cx = 0
		cy = 0
		cthick = 2
		lthick = 1
		labarm = 0.1
		labspc = 0.01
		minres = pi*2*100
	}
	
	function error(msg)
	{
		print "animpie gen_pie(): " msg > "/dev/stderr"
		fatal=1
		exit 1
	}

	/@cx/     { getline cx; next }
	/@cy/     { getline cy; next }
	/@radius/ { getline radius; next }
	/@minres/ { getline minres; next }
	/@cthick/ { getline cthick; next }
	/@lthick/ { getline lthick; next }
	/@labarm/ { getline labarm; next }
	/@labspc/ { getline labspc; next }
	/@total/  { getline total; next }

	/@slice/  { slice=int(slices); slices++; getline SIZE[slice]; next }
	/@label/  { getline LABEL[slice]; next }
	/@color/  { getline COLOR[slice]; next }
	/@pull/   { getline PULL[slice]; next }
	{
		line=$0
		sub("#.*", "", line)
		if ($0 ~ "^[ \t]*")
			next
		error("Syntax error in line " NR)
	}

	# determine wildcard sizes
	# sets global SIZE[n], sum
	function sum_and_wildcard(    n,wild_size,diff)
	{
		for(n = 0; n < slices; n++) {
			if (SIZE[n] != "*")
				sum = sum + (SIZE[n] + 0.0)
			else
				wild++
		}
		if (wild > 0) {
			if (total == "")
				error("wildcard slice size works only if total is specified")
			wild_size = (total - sum) / wild
			for(n = 0; n < slices; n++) {
				if (SIZE[n] == "*") {
					SIZE[n] = wild_size
					sum += wild_size
				}
			}
		}
		if (total != "") {
			diff = sum - total
			if (diff < 0) diff = -diff
			if (diff > total/100)
				error("sum of slice sizes mismatch total size specified")
		}
		total = sum
		smallest = sum
		for(n = 0; n < slices; n++)
			if (SIZE[n] < smallest)
				smallest = SIZE[n]
	}

	function rsin(a) { return sin(a+pi/2) }
	function rcos(a) { return cos(a+pi/2) }
	function abs(a)  { return (a > 0) ? a : -a }

	# Generate coordinates for slice contours and label startpoint at poly centerline
	# sets POLY[] and POLYCLP[] and POLYANG[]
	function gen_poly(    ang,n,s,astep,d_ang,t_ang,start,pcx,pcy,res)
	{
		res = (total/smallest*2)
		astep = 2*pi / ((res < minres) ? minres : res)
		for(n = 0; n < slices; n++) {
			d_ang = (SIZE[n] / total) * 2*pi
			if (d_ang < 2*astep)
				d_ang = 2*astep
			a = ang + d_ang/2
			POLYANG[n] = a
			pcx = cx + PULL[n] * rcos(a)*radius
			pcy = cy + PULL[n] * rsin(a)*radius
			POLYCLP[n, "x"] = pcx+rcos(a)*(radius*0.9)
			POLYCLP[n, "y"] = pcy+rsin(a)*(radius*0.9)
			start = pcx " " pcy
			s = start
			if (n == slices - 1)
				t_ang = 2*pi
			else
				t_ang = ang+d_ang
			for(1; ang < t_ang; ang += astep)
				s = s "  " pcx+rcos(ang)*radius " " pcy+rsin(ang)*radius
			ang -= astep
			s = s "  " start
			POLY[n] = s
		}
	}

	function draw_label(n, dx,   ext,c,s,ky)
	{
		if (LABEL[n] == 0)
			return

		ext=0.2
		c = rcos(POLYANG[n])*radius*ext
		s = rsin(POLYANG[n])*radius*ext
		ky = POLYCLP[n, "y"]

		do {
			ky += s
		} while(abs(ky - lastky) < 0.05)

		print "lines", POLYCLP[n, "x"], POLYCLP[n, "y"], \
		      POLYCLP[n, "x"] + c, ky, \
		      POLYCLP[n, "x"] + c + dx*labarm, ky
		print "text ", \
		      (dx < 0) ? "left" : "right", "center", \
		      POLYCLP[n, "x"] + c + dx*(labarm+labspc), ky, "\"" LABEL[n] "\""
		lastky = ky
	}

	function draw(    n,dx)
	{
		# draw polygons
		if (cthick != 0)
			print "thick " cthick
		for(n = 0; n < slices; n++) {
			if (COLOR[n] == "")
				print "color white"
			else
				print "color " COLOR[n]
			print "poly " POLY[n]
			if (cthick != 0) {
				print "color black"
				print "lines " POLY[n]
			}
		}

		# draw labels
		if (lthick != 0)
			print "thick " lthick

		lastky=1000
		# left side, from top to bottom
		for(n = 0; n < slices; n++)
			if ((POLYANG[n] >= 0) && (POLYANG[n] <= pi))
				draw_label(n, -1)
		# right side, from top to bottom
		for(n = slices-1; n >= 0; n--)
			if (!((POLYANG[n] >= 0) && (POLYANG[n] <= pi)))
				draw_label(n, 1)

	}

	END {
		if (fatal)
			exit
		sum_and_wildcard()
		gen_poly()
		if (!embed)
			print "frame"
		draw()
		if (!embed)
			print "flush"
	}
'



